(* Title:  BaseDefinitions/theoryC.sml
   Author: Walther Neuper
   (c) due to copyright terms

Probably this structure can be dropped due to improved reflection in Isabelle.
*)
signature THEORY_ISAC =
sig
  eqtype id
  val cut_id: string -> id
  val id_of: theory -> id
  val curried_equal: theory -> theory -> bool

  val last_isabelle_thy: theory
  val session_specify_names: id list
  val session_interpret_names: id list
  val get_theory: id -> theory         (* restricted to Isac_Knowledge *)
  val get_theory_PIDE: Proof.context -> id -> theory
  val id_to_ctxt: id -> Proof.context

  val id_empty: id
  val Isac: 'a -> theory
  val parent_of: theory -> theory -> theory

  type metadata
  val the_metadata: theory -> metadata
  val set_metadata: metadata -> theory -> theory

\<^isac_test>\<open>
  val get_thy: id -> theory            (* restricted to session Isac *)
\<close>
end

(**)
structure ThyC(**): THEORY_ISAC(**) =
struct
(**)

type id = string;
val id_empty = "empty_thy_id";
fun cut_id dn = hd (space_explode "." dn);
fun id_of thy = thy |> Context.theory_id |> Context.theory_id_name
fun curried_equal thy1 thy2 = id_of thy1 = id_of thy2

val last_isabelle_thy = @{theory Complex_Main}
val session_specify_names = ["Specify", "MathEngBasic", "Input_Descript",
  "ProgLang", "Auto_Prog", "Prog_Expr", "Tactical", "Prog_Tac", "Program", "ListC", "Calc_Binop",
  "BaseDefinitions", "Know_Store"]
val session_interpret_names = ["Interpret"]
(*this is troublesome since the introduction of sessions*)
fun get_thy thyID =
  if member op = session_specify_names thyID
  then Thy_Info.get_theory ("Specify." ^ thyID)
  else if member op = session_interpret_names thyID
  then Thy_Info.get_theory ("Interpret." ^ thyID)
  else Thy_Info.get_theory ("Isac." ^ thyID);
fun get_theory_PIDE ctxt thy_id =
  let
    val known_thys = Theory.nodes_of (Proof_Context.theory_of ctxt)
  in 
    (find_first (fn thy => Context.theory_name thy = thy_id) known_thys
      |> the)
    handle Option.Option => raise ERROR ("get_theory fails with " ^ quote thy_id)
  end
fun get_theory thy =   
    if thy = id_empty
    then (get_thy "Base_Tools") (*lower bound of Knowledge*)
    else (get_thy thy) handle ERROR _ => raise ERROR ("ME_Isa: thy \"" ^ thy ^ "\" not in system");

fun id_to_ctxt thy' = Proof_Context.init_global (get_theory thy');

fun Isac _ = 
 (get_theory "Isac_Knowledge");

fun parent_of thy1 thy2 = if Context.subthy (thy1, thy2) then thy2 else thy1;


(** meta data **)

type authors = string list; (* no more for thm, rls, rew_ord, cal *)
type metadata = {coursedesign: authors, mathauthors: authors}
(*WAS IN Thy_Write:  datatype thydata
    = ...
    | Html of {coursedesign: authors, guh: Check_Unique.id, html: string, mathauthors: authors}*)
val default_metadata = {coursedesign = ["(c) isac team 2006"], mathauthors = []}

structure Data = Theory_Data
(
  type T = metadata option;
  val empty = SOME default_metadata;
  fun merge _ = NONE;
);
val the_metadata = the o Data.get;
val set_metadata = Data.put o SOME;


(**)end(**)
