(*.(c) by Richard Lang, 2003 .*)
(* theory collecting all knowledge for RationalEquations
   created by: rlang 
         date: 02.08.12
   changed by: rlang
   last change by: rlang
             date: 02.11.28
*)

theory RatEq imports Rational LinEq begin

text \<open>univariate equations over multivariate rational terms:
  In 2003 this type has been integrated into ISAC's equation solver
  by Richard Lang; the root for the solver is Equation.thy.
  The migration Isabelle2002 --> 2011 found that application of theorems like 
  rat_mult_denominator_right: "[|Not(d=0)|] ==> ((a::real) = c / d) = (a*d = c)"
  in rule-sets does not transfer "d ~= 0" to the assumptions; see
  test --- repair NO asms from rls RatEq_eliminate ---.
  Thus the migration dropped update of Check_elementwise, which would require
  these assumptions; see 
  test --- pbl: rational, univariate, equation ---, --- x / (x ^ 2 - 6 * x + 9) - 1 /...
\<close>

subsection \<open>consts definition for predicates in specifications\<close>
consts
  is_ratequation_in :: "[bool, real] => bool" ("_ is'_ratequation'_in _")

subsection \<open>theorems not yet adopted from Isabelle\<close>
axiomatization where
   (* FIXME also in Poly.thy def. --> FIXED*)
   (*real_diff_minus            
   "a - b = a + (-1) * b"*)
   real_rat_mult_1:   "a*(b/c) = (a*b)/c" and
   real_rat_mult_2:   "(a/b)*(c/d) = (a*c)/(b*d)" and
   real_rat_mult_3:   "(a/b)*c = (a*c)/b" and
   real_rat_pow:      "(a/b) \<up> 2 = a \<up> 2/b \<up> 2" and

   rat_double_rat_1:   "[|Not(c=0); Not(d=0)|] ==> (a / (c/d) = (a*d) / c)" and
   rat_double_rat_2:   "[|Not(b=0);Not(c=0); Not(d=0)|] ==> 
                                           ((a/b) / (c/d) = (a*d) / (b*c))" and
   rat_double_rat_3:   "[|Not(b=0);Not(c=0)|] ==> ((a/b) / c = a / (b*c))" and

  (* equation to same denominator *)
  rat_mult_denominator_both:
   "[|Not(b=0); Not(d=0)|] ==> ((a::real) / b = c / d) = (a*d = c*b)" and
  rat_mult_denominator_left:
   "[|Not(d=0)|] ==> ((a::real) = c / d) = (a*d = c)" and
  rat_mult_denominator_right:
   "[|Not(b=0)|] ==> ((a::real) / b = c) = (a = c*b)"

subsection \<open>predicates\<close>
ML \<open>
(*-------------------------functions-----------------------*)
(* is_rateqation_in becomes true, if a bdv is in the denominator of a fraction*)
fun is_rateqation_in t v = 
  let 
   	fun finddivide (t as (_ $ _ $ _ $ _)) _ = raise TERM ("is_rateqation_in", [t]) 
	    (* at the moment there is no term like this, but ....*)
	  | finddivide (Const (\<^const_name>\<open>divide\<close>,_) $ _ $ b) v = Prog_Expr.occurs_in v b
	  | finddivide (_ $ t1 $ t2) v = finddivide t1 v orelse finddivide t2 v
	  | finddivide (_ $ t1) v = finddivide t1 v
	  | finddivide _ _ = false;
  in
	  finddivide t v
  end;
\<close>

subsection \<open>evaluations functions\<close>
ML \<open>
fun eval_is_ratequation_in _ _ (p as (Const (\<^const_name>\<open>RatEq.is_ratequation_in\<close>, _) $ t $ v)) _ =
    if is_rateqation_in t v
    then SOME ((UnparseC.term p) ^ " = True", HOLogic.Trueprop $ (TermC.mk_equality (p, @{term True})))
    else SOME ((UnparseC.term p) ^ " = True", HOLogic.Trueprop $ (TermC.mk_equality (p, @{term False})))
  | eval_is_ratequation_in _ _ _ _ = ((*tracing"### nichts matcht";*) NONE);
\<close>
calculation is_ratequation_in = \<open>eval_is_ratequation_in ""\<close>

subsection \<open>rule-sets\<close>
ML \<open>
val RatEq_prls = (*15.10.02:just the following order due to subterm evaluation*)
  Rule_Set.append_rules "RatEq_prls" Rule_Set.empty [
    \<^rule_eval>\<open>Prog_Expr.ident\<close> (Prog_Expr.eval_ident "#ident_"),
    \<^rule_eval>\<open>Prog_Expr.matches\<close> (Prog_Expr.eval_matches ""),
    \<^rule_eval>\<open>Prog_Expr.lhs\<close> (Prog_Expr.eval_lhs ""),
    \<^rule_eval>\<open>Prog_Expr.rhs\<close> (Prog_Expr.eval_rhs ""),
    \<^rule_eval>\<open>is_ratequation_in\<close> (eval_is_ratequation_in ""),
    \<^rule_eval>\<open>HOL.eq\<close> (Prog_Expr.eval_equal "#equal_"),
    \<^rule_thm>\<open>not_true\<close>,
    \<^rule_thm>\<open>not_false\<close>,
    \<^rule_thm>\<open>and_true\<close>,
    \<^rule_thm>\<open>and_false\<close>,
    \<^rule_thm>\<open>or_true\<close>,
    \<^rule_thm>\<open>or_false\<close>];

\<close> ML \<open>
(*rls = Rule_Set.merge asm_rls Poly_erls *)
val rateq_erls = 
  Rule_Set.keep_unique_rules "rateq_erls"                             (*WN: ein Hack*)
	  (Rule_Set.merge "is_ratequation_in" calculate_Rational
		  (Rule_Set.append_rules "is_ratequation_in" Poly_erls [
         (*\<^rule_eval>\<open>divide\<close> (Prog_Expr.eval_cancel "#divide_e"),*)
			    \<^rule_eval>\<open>is_ratequation_in\<close> (eval_is_ratequation_in "")])) [
      \<^rule_thm>\<open>and_commute\<close>, (*WN: ein Hack*)
    	\<^rule_thm>\<open>or_commute\<close>];  (*WN: ein Hack*)

\<close> ML \<open>
val RatEq_crls = 
  Rule_Set.keep_unique_rules "RatEq_crls"                              (*WN: ein Hack*)
	  (Rule_Set.merge "is_ratequation_in" calculate_Rational
		  (Rule_Set.append_rules "is_ratequation_in" Poly_erls [
         (*\<^rule_eval>\<open>divide\<close> (Prog_Expr.eval_cancel "#divide_e"),*)
		    \<^rule_eval>\<open>is_ratequation_in\<close> (eval_is_ratequation_in "")])) [
      \<^rule_thm>\<open>and_commute\<close>, (*WN: ein Hack*)
    	\<^rule_thm>\<open>or_commute\<close>];  (*WN: ein Hack*)

val RatEq_eliminate = prep_rls'(
  Rule_Def.Repeat {
    id = "RatEq_eliminate", preconds = [], rew_ord = ("termlessI", termlessI), asm_rls = rateq_erls,
    prog_rls = Rule_Set.Empty, calc = [], errpatts = [],
    rules = [
     \<^rule_thm>\<open>rat_mult_denominator_both\<close>, (* a/b=c/d -> ad=cb *)
     \<^rule_thm>\<open>rat_mult_denominator_left\<close>, (* a  =c/d -> ad=c  *)
     \<^rule_thm>\<open>rat_mult_denominator_right\<close>],  (* a/b=c   ->  a=cb *)
   program = Rule.Empty_Prog});

\<close> ML \<open>
val RatEq_simplify = prep_rls'(
  Rule_Def.Repeat {
    id = "RatEq_simplify", preconds = [], rew_ord = ("termlessI", termlessI), asm_rls = rateq_erls,
    prog_rls = Rule_Set.Empty, calc = [], errpatts = [],
    rules = [
      \<^rule_thm>\<open>real_rat_mult_1\<close>, (*a*(b/c) = (a*b)/c*)
      \<^rule_thm>\<open>real_rat_mult_2\<close>, (*(a/b)*(c/d) = (a*c)/(b*d)*)
      \<^rule_thm>\<open>real_rat_mult_3\<close>, (* (a/b)*c = (a*c)/b*)
      \<^rule_thm>\<open>real_rat_pow\<close>, (*(a/b) \<up> 2 = a \<up> 2/b \<up> 2*)
      \<^rule_thm>\<open>real_diff_minus\<close>, (* a - b = a + (-1) * b *)
      \<^rule_thm>\<open>rat_double_rat_1\<close>, (* (a / (c/d) = (a*d) / c) *)
      \<^rule_thm>\<open>rat_double_rat_2\<close>, (* ((a/b) / (c/d) = (a*d) / (b*c)) *)
      \<^rule_thm>\<open>rat_double_rat_3\<close> (* ((a/b) / c = a / (b*c) ) *)],
    program = Rule.Empty_Prog});
\<close>
rule_set_knowledge rateq_erls = rateq_erls
rule_set_knowledge RatEq_eliminate = RatEq_eliminate
rule_set_knowledge RatEq_simplify = RatEq_simplify

subsection \<open>problems\<close>

problem pbl_equ_univ_rat : "rational/univariate/equation" =
  \<open>RatEq_prls\<close>
  Method_Ref: "RatEq/solve_rat_equation"
  CAS: "solve (e_e::bool, v_v)"
  Given: "equality e_e" "solveFor v_v"
  Where: "(e_e::bool) is_ratequation_in (v_v::real)"
  Find: "solutions v_v'i'"

subsection \<open>methods\<close>

method met_rateq : "RatEq" =
  \<open>{rew_ord="tless_true",rls'=Atools_erls,calc = [], prog_rls = Rule_Set.empty, where_rls=Rule_Set.empty,
    errpats = [], rew_rls = norm_Rational}\<close>

partial_function (tailrec) solve_rational_equ :: "bool \<Rightarrow> real \<Rightarrow> bool list"
  where
"solve_rational_equ e_e v_v =
  (let
    e_e = (
      (Repeat (Try (Rewrite_Set ''RatEq_simplify''))) #>
      (Repeat (Try (Rewrite_Set ''norm_Rational''))) #>
      (Repeat (Try (Rewrite_Set ''add_fractions_p''))) #>
      (Repeat (Try (Rewrite_Set ''RatEq_eliminate''))) ) e_e;
    L_L = SubProblem (''RatEq'', [''univariate'', ''equation''], [''no_met'']) [BOOL e_e, REAL v_v]
  in
    Check_elementwise L_L {(v_v::real). Assumptions})"

method met_rat_eq : "RatEq/solve_rat_equation" =
  \<open>{rew_ord="termlessI", rls'=rateq_erls, prog_rls=Rule_Set.empty, where_rls=RatEq_prls, calc=[],
    errpats = [], rew_rls = norm_Rational}\<close>
  Program: solve_rational_equ.simps
  Given: "equality e_e" "solveFor v_v"
  Where: "(e_e::bool) is_ratequation_in (v_v::real)"
  Find: "solutions v_v'i'"

ML \<open>
\<close> ML \<open>
\<close> ML \<open>
\<close> 
end
