(* Title: positions in the calctree
   Author: Walther Neuper 2019
   (c) due to copyright terms
*)

signature POSITION =
sig
  eqtype posel
  type pos = posel list
  val pos2str: int list -> string
  datatype pos_ = Frm | Met | Pbl | Res | Und
  val pos_2str: pos_ -> string

(*type T*)
  type pos'

  val pos'2str: pos' -> string
  val str2pos_: string -> pos_
  val e_pos': pos'

(*/------------------- for hiding constructors ---------------------------------------------\*)
  val on_calc_end: pos' -> bool

(*val at_first_tactic: term -> TermC.path -> pos' -> bool*)
  val at_first_tactic: pos' -> bool
(*\------------------- for hiding constructors ---------------------------------------------/*)
  val lev_of: pos' -> int
  val lev_dn: pos -> pos                       (* duplicate in ctree-basic.sml *)
  val lev_dn_: pos' -> pos'
  val lev_on : pos -> pos                      (* duplicate in ctree-basic.sml *)
  val lev_on_total : pos -> pos                      (* duplicate in ctree-basic.sml *)
  val pos_plus : int -> pos' -> pos'
  val lev_up : pos -> pos
  val lev_back' : pos' -> pos'                                (* for inform.sml *)
  val lev_back : pos' -> pos'                                 (* for inform.sml *)
  val start_new_level: pos' -> pos'
  val back_from_new: pos' -> pos'
  val new_on_level: pos' -> bool
  val next_in_prog': pos' -> pos'
  val next_in_prog: pos' -> pos
  val on_specification : pos' -> bool
  val on_specification': pos' -> bool

\<^isac_test>\<open>
  val pos's2str: pos' list -> string
\<close>
end

(**)
structure Pos(**): POSITION(**) =
struct
(**)
type posel = int; (* for readability in funs accessing Ctree *)
type pos = int list;
val pos2str = ints2str';
datatype pos_ = 
  Pbl    (* PblObj-position: problem-type                   *)
| Met    (* PblObj-position: method                         *)
| Frm    (* PblObj-position: -> Pbl in ME (not by moveDown !)
         |  PrfObj-position: formula                        *)
| Res    (* PblObj | PrfObj-position: result                *)
| Und;   (* undefined*)
fun pos_2str Pbl = "Pbl"
  | pos_2str Met = "Met"
  | pos_2str Frm = "Frm"
  | pos_2str Res = "Res"
  | pos_2str Und = "Und";
fun str2pos_ "Pbl" = Pbl
  | str2pos_ "Met" = Met
  | str2pos_ "Frm" = Frm
  | str2pos_ "Res" = Res
  | str2pos_ "Und" = Und
  | str2pos_ str = raise ERROR ("str2pos_: wrong argument = " ^ str)

type pos' = pos * pos_;
(*WN0312 remembering interator (pos * pos_) for ctree 
	   pos : lev_on, lev_dn, lev_up
     pos_:
# generate1 sets pos_ if possible  ...?WN0502?NOT...
# generate1 does NOT set pos, because certain nodes can be lev_on OR lev_dn
                     exceptions: Begin/End_Trans
# thus generate(1) called in
.# scan_dn1, locate_input_tactic 
.# by_tactic (tac_ -cases); general case: 
  val pos' = case pos' of (p,Res) => (lev_on p',Res) | _ => pos'
# WN050220, S(604):
  generate1...(Rewrite(f,..,res))..(pos, pos_)
     cappend_atomic.................pos //////  gets f+res always!!!
        cut_tree....................pos, pos_ 
*)
fun pos'2str (p, p_) = pair2str (ints2str' p, pos_2str p_);
\<^isac_test>\<open>
fun pos's2str ps = (strs2str' o (map pos'2str)) ps;
\<close>
val e_pos' = ([], Und);

fun on_calc_end ([], Res) = true
  | on_calc_end _ = false;

fun at_first_tactic (p, p_) = last_elem p = 0 andalso p_ = Res
(*
fun at_first_tactic prog path (pos as (p, p_)) =
  (if not (path = []) andalso (last_elem p = 0 andalso p_ = Res)
    then raise ERROR ("at_start_detail: not (path = []) andalso at_start_detail in \n\""
      ^ UnparseC.term prog ^ "\"\n at \"" ^ TermC.string_of_path path ^ "\" for \"" ^ pos'2str pos ^ "\"")
    else ();
  last_elem p = 0 andalso p_ = Res)
*)

fun lev_of (p,_) = length p;
fun lev_dn p = p @ [0];
fun lev_dn_ (p, _) = (lev_dn p, Res)
(* duplicates in ctree-basic.sml *)
fun lev_on [] = raise ERROR "lev_on []"
  | lev_on pos = 
    let val len = length pos
    in (drop_last pos) @ [(nth len pos)+1] end;
fun lev_on_total [] = [0]
  | lev_on_total pos = 
    let val len = length pos
    in (drop_last pos) @ [(nth len pos)+1] end;
(* increase pos by n within a level *)
fun pos_plus 0 pos = pos
  | pos_plus n (p, Frm) = pos_plus (n - 1) (p, Res)
  | pos_plus n (p, _) = pos_plus (n - 1) (lev_on p, Res);
fun lev_up [] = raise ERROR "lev_up []"
  | lev_up p = (drop_last p):pos;
(*040216: for inform --> embed_deriv: remains on same level TODO.WN120517 compare lev_pred*)
fun lev_back' ([], _) = raise ERROR "lev_back': called by ([],_)"
  | lev_back' (p, _) =
    if last_elem p <= 1 then (p, Frm) 
    else ((drop_last p) @ [(nth (length p) p) - 1], Res);
fun lev_back ([], p_) = ([], p_)
  | lev_back (p, _) =
    if last_elem p <= 1 then (p, Frm) 
    else ((drop_last p) @ [(nth (length p) p) - 1], Res);
fun start_new_level (p, _) = ((lev_on o lev_dn) p, Frm)
fun new_on_level (p, Frm) =
      if last_elem p = 1 then true else false
  | new_on_level _ = false
fun back_from_new (p, p_) =
  let
    val p' = if last_elem p = 0 then drop_last p else p
  in (p', p_) end

fun next_in_prog' (pos as (_, Met)) = start_new_level pos
 	| next_in_prog' (p, Res) = (lev_on p, Res)
 	| next_in_prog' pos = pos
fun next_in_prog (p, Frm) = p
  | next_in_prog (p, Res) = lev_on p
  | next_in_prog pos = raise ERROR ("Step_Solve.by_tactic: call by " ^ pos'2str pos);

(*WN.12.5.03: quick-and-dirty repair for listexpressions*)
fun on_specification (_, Pbl) = true
  | on_specification (_, Met) = true
  | on_specification _ = false;
fun on_specification' p = on_specification p orelse new_on_level p;

(**)end(**)
