(* Quick introduction to ML, session 3 *)

theory ML3_Combinators imports Isac.Isac_Knowledge begin

text\<open>This course follows the book:
J.D.Ullmann, Elements of ML Programming. Prenctice Hall 1998.
Contents:\<close>

section \<open>5.6 Higher-Order Functions\<close>

subsection \<open>map\<close>ML \<open>
map;
fun add3 a = a + 3: int;
map add3 [];
map add3 [1,2,3,4,5];
\<close>
ML \<open>
fun greater a b = a > b;
greater 2 3;
greater 5 3;
greater 5;
map (greater 3) [1,2,3,4,5];
\<close>

subsection \<open>fold\<close>
ML \<open>
fold;
fun add a b = a + b;
fun mult a b = a * b;
\<close>
ML \<open>
fold add [1,2,3,4,5] 10;
fold add [1,2,3,4,5] 10;

fold mult [1,2,3] 1;
fold mult [1,2,3,4,5] 1;
\<close>

section \<open>Bad style: local variables in functions\<close>
ML \<open>
fun xxx a b c d e f =
  let 
    val x = a + b
    val _ = writeln ("### xxx: x = " ^ string_of_int x)
    val y = c * d
    val z = e ^ f
  in (x, y , z) end;
xxx 1 2 3 4  "a" "b";
\<close>
text \<open>little improvement\<close>
ML \<open>
fun xxx a b c d e f =
  let 
    val x = a + b
    val _ = writeln (string_of_int x)
    val _ = string_of_int x |> writeln
    val y = c * d
    val z = e ^ f
  in (x, y , z) end;
xxx 1 2 3 4  "a" "b";
\<close>


section \<open>Combinators, see src/Pure/General/basics.ML\<close>
subsection \<open>Preliminaries\<close>
ML \<open>
(*infix +++;*)
fun +++ a b = a + b;
+++ 1 3;
\<close>
ML \<open>
fn x => x + 10; (* function without a name*)
(fn x => x + 10) 3;

fun ??? x = x + 10;
\<close>

subsection \<open>Combinators with example in test/Pure/General/Basics.thy\<close>
text \<open>see test/Pure/General/Basics.thy\<close>
ML \<open>
op |> 
\<close>
ML \<open>
op ||> ;
\<close>

subsection \<open>Still no examples in test/Pure/General/Basics.thy\<close>
text \<open>please, find examples for ||>> ff !\<close>
ML \<open>
op #>>;
fun fff a = (a + 10, string_of_int a);
fun ggg b = b * 2;
fff #>> ggg;
(fff #>> ggg) 1;
\<close>

end
